import igraph as ig
import random
import numpy as np
from collections import Counter

def generate_family_edges(G):
    """
    Generate family edges. All edges between nodes of the same family are created
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network
        
    Return
    ------
    None

    """
    toAdd = []

    for node in G.vs:
        for edge in node["family_contacts"]:
            if G.vs[edge[0]]["agent_status"] != "D" \
            and G.vs[edge[1]]["agent_status"] != "D" \
            and not G.vs[edge[0]]["quarantine"] \
            and not G.vs[edge[1]]["quarantine"]:
            #and not (edge[0], edge[1]) in toAdd
            #and not (edge[1], edge[0]) in toAdd:
                toAdd.append(edge)

    weights = np.random.randint(3, 8, len(toAdd))
    G.add_edges(toAdd)
    for (edge_index, i) in zip(G.get_eids(toAdd, directed = False), range(len(weights))):
      G.es[edge_index]["weight"] = weights[i]
      G.es[edge_index]["category"] = "family_contacts"  

def generate_occfreq_edges(G, edge_category, restriction_value):
    """
    Create edges from the node of type edge_category
    The number of edges is chosen according to the sociality of the node
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network

    edge_category: string
        Category of the edge attribute
    
    restriction_value: float
        How many edges are dropped in proportion to normal condition?
        
    Return
    ------
    None

    """
    toAdd = []

    for node in G.vs:
        possible_edges = node[edge_category].copy()
        random.shuffle(possible_edges)
        tmp = int(len(possible_edges) / 3)
        tmp2 = int(2 * len(possible_edges) / 3)
        if node["sociality"] == "low":
            n_edges = int(random.random() * tmp)
        if node["sociality"] == "medium":
            n_edges = tmp + int(random.random() * (tmp2 - tmp))
        if node["sociality"] == "high":
            n_edges = tmp2 + int(random.random() * (len(possible_edges) + 1 - tmp2))
        n_edges = int(n_edges * restriction_value)

        for i in range (0, n_edges):
            edge = possible_edges[i]
            if G.vs[edge[0]]["agent_status"] != "D" \
            and G.vs[edge[1]]["agent_status"] != "D" \
            and not G.vs[edge[0]]["quarantine"] \
            and not G.vs[edge[1]]["quarantine"] \
            and not G[edge[0], edge[1]]:
            #and not (edge[0], edge[1]) in toAdd \
            #and not (edge[1], edge[0]) in toAdd:
                toAdd.append(edge)

    weights = np.random.randint(1, 6, len(toAdd))
    G.add_edges(toAdd)
    for (edge_index, i) in zip(G.get_eids(toAdd, directed = False), range(len(weights))):
      G.es[edge_index]["weight"] = weights[i]
      G.es[edge_index]["category"]= edge_category  

def generate_random_edges(G, number_of_random_edges, restriction_value):
    """
    Create number_of_random_edges random edges in the contact network
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network
        
    number_of_random_edges : int
        Number of edges to create
    
    restriction_value: float
        How many edges are dropped in proportion to normal condition?
        
    Return
    ------
    None

    """
    toAdd = []
    number_of_random_edges = int(number_of_random_edges * restriction_value)
    edge_list = np.random.randint(0, len(list(G.vs)) - 1, 2 * number_of_random_edges)
    for i in range (0, 2 * number_of_random_edges, 2):
        source = edge_list[i]
        target = edge_list[i + 1]
        if not G[source, target] \
        and source != target \
        and G.vs[source]["agent_status"] != "D" \
        and G.vs[target]["agent_status"] != "D" \
        and not G.vs[source]["quarantine"] \
        and not G.vs[target]["quarantine"] \
        and not G[source, target]:
        #and not (source, target) in toAdd \
        #and not (target, source) in toAdd:
            toAdd.append((source, target))

    G.add_edges(toAdd)
    for edge_index in G.get_eids(toAdd, directed = False):
      G.es[edge_index]["weight"] = 1
      G.es[edge_index]["category"]= "random_contacts"  

def step_edges(G, restriction_value):
    """
    Remove old edges
    Create new edges
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network

    restriction_value: float
        How many edges are dropped in proportion to normal condition?
        
    Return
    ------
    None

    """
    
    G.delete_edges(list(G.es))

    generate_family_edges(G)

    generate_occfreq_edges(G, "frequent_contacts", restriction_value)
    generate_occfreq_edges(G, "occasional_contacts", restriction_value)

    random_contact_total = len(list(G.vs)) + random.random() * (7 * len(list(G.vs)) - len(list(G.vs)))
    generate_random_edges(G, random_contact_total, restriction_value)

    # since edge generation produce a multigraph but a single edge has attributes as wanted,
    # all other edges are removed. This produce a simple graph
    # this operation is way much faster than checking befor adding new edge
    
    toRemove = list()
    for edge in G.es:
        if edge["weight"] == None:
            toRemove.append(edge)
    G.delete_edges(toRemove)  

def step_spread(G, incubation_days, infection_duration, infection_rate):
    """
    Make the infection spread across the network
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network

    incubation_days: int
        Average number of days where the patient is not infective

    infection_duration: int
        Average total duration of the disease
    
    infection_rate: float
        Value of the infection_rate rate for the disease in the network

    Return
    ------
    None
    
    """
    for node in G.vs:
        # update parameters if node is infected
        if node["infected"] == True:
            node["days_from_infection"] += 1
            if node["days_from_infection"] == infection_duration:
                # if infection is over, it will be dead of recovered
                node["agent_status"] = np.random.choice(["R", "D"], p = (1 - node["death_rate"], node["death_rate"]))
                node["infected"] = False
                node["symptoms"] = list()
                node["days_from_infection"] = 0
                node["needs_IC"] = False
                if node["agent_status"] == "D":
                    node["quarantine"] = 0
                    node["test_result"] = - 1
                
            # if it is still infective, spread the infection with his contacts
            if node["agent_status"] == "I":
                for contact in G.neighborhood(node)[1:]:
                    if G.vs[contact]["agent_status"] == "S":
                        prob = infection_rate * G[node, contact] # access to the weight of the edge
                        # has the new node been infected?
                        G.vs[contact]["agent_status"] = np.random.choice(["S", "E"], p = (1 - prob, prob))
                        if G.vs[contact]["agent_status"] == "E":
                            G.vs[contact]["infected"] = True
                            G.vs[contact]["days_from_infection"] = 1

            # if the node become I, pick some symptoms
            if node["agent_status"] == "E" and node["days_from_infection"] == incubation_days:
                node["agent_status"] = "I"
                #if mild case
                case = random.uniform(0, 1)
                if case < 0.8:
                    if case < 0.05:
                        node["symptoms"].append("Loss of taste or smell")
                    if case < 0.2:
                        node["symptoms"].append("Fever")
                    if case < 0.2:
                        node["symptoms"].append("Cough")
                    if case < 0.2:
                        node["symptoms"].append("Tiredness")
                # if severe
                else:
                    if case < 0.99:
                        node["symptoms"].append("Fever")
                    if case < 0.7:
                        node["symptoms"].append("Tiredness")
                    if case < 0.6:
                        node["symptoms"].append("Cough")
                    if case < 0.3:
                        node["symptoms"].append("Dyspnea")
                
                if random.uniform(0, 1) < 0.02:
                    node["needs_IC"] = True
   
def step_test(G, nets, incubation_days, n_new_test, policy_test, contact_tracking_efficiency):
    """
    Test some nodes of the network and put the in quarantine if needed
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network

    nets: list of ig.Graph()
        History of the network
    
    incubation_days: int
        Average number of days where the patient is not infective

    n_new_test: int
        Number of new avaiable tests

    policy_test: string
        Test strategy
        Can be ["Random, Degree Centrality, Betweenness Centrality"]
    
    contact_tracking_efficiency: float
        The percentage of contacts successfully traced

    Return
    ------
    None

    """

    # create pool of nodes to test
    high_priority_test_pool = set()
    low_priority_test_pool = set()
    for node in G.vs:
        # update quarantine
        if node["quarantine"] > 0:
            node["quarantine"] -= 1
            # if node has been found positive and quarantine is over, re-test the node
            if node["quarantine"] == 0 and node["test_result"] == 1:
                high_priority_test_pool.add(node.index)
        # update test validity
        if node["test_validity"] > 0:
            node["test_validity"] -= 1
        # test also detect if the node is recovered (sierological test)
        # if node is not dead, if test validity is expired, if node is not a known recovered, add to low priority test pool
        if node["agent_status"] != "D" and node["test_validity"] <= 0 \
         and not (node["test_result"] == 0 and node["agent_status"] == "R"):
            low_priority_test_pool.add(node.index)

    low_priority_test_pool = low_priority_test_pool - high_priority_test_pool
    found_positive = set()
    #cannot directly put nodes in a set cause node object are not supported
    high_priority_test_pool = [G.vs[i] for i in high_priority_test_pool]
    low_priority_test_pool = [G.vs[i] for i in low_priority_test_pool]
    for node in high_priority_test_pool:
        if node["infected"]:
            node["test_result"] = 1
            node["quarantine"] = 14
            node["test_validity"] = 14
            found_positive.add(node.index)
        else:
            node["test_result"] = 0
            node["test_validity"] = incubation_days

    if policy_test == "Random" and n_new_test > 0:
        to_test = random.sample(low_priority_test_pool, min(len(low_priority_test_pool), n_new_test))
        for node in to_test:
            if node["infected"]:
                node["test_result"] = 1
                node["quarantine"] = 14
                node["test_validity"] = 14
                found_positive.add(node.index)
            else:
                node["test_result"] = 0
                node["test_validity"] = incubation_days

    if policy_test == "Degree Centrality" and n_new_test > 0:
        low_priority_test_pool_index = [x.index for x in low_priority_test_pool]
        degree_results = G.strength(low_priority_test_pool_index, weights = "weight")
        zipped_lists = zip(degree_results, low_priority_test_pool)
        sorted_pairs = sorted(zipped_lists, reverse = True)
        tuples = zip(*sorted_pairs)
        _, sorted_nodes = [ list(tuple) for tuple in  tuples]
        for i in range(min(n_new_test, len(low_priority_test_pool))):
            node = sorted_nodes[i]
            if node["infected"]:
                node["test_result"] = 1
                node["quarantine"] = 14
                node["test_validity"] = 14
                found_positive.add(node.index)
            else:
                node["test_result"] = 0
                node["test_validity"] = incubation_days

    if policy_test == "Betweenness Centrality" and n_new_test > 0:
        low_priority_test_pool_index = [x.index for x in low_priority_test_pool]
        betweenness_results = G.betweenness(low_priority_test_pool_index, 
                                            directed = False, weights = "weight",
                                            cutoff = None)
        zipped_lists = zip(betweenness_results, low_priority_test_pool)
        sorted_pairs = sorted(zipped_lists, reverse = True)
        tuples = zip(*sorted_pairs)
        _, sorted_nodes = [ list(tuple) for tuple in  tuples]
        for i in range(min(n_new_test, len(low_priority_test_pool))):
            node = sorted_nodes[i]
            if node["infected"]:
                node["test_result"] = 1
                node["quarantine"] = 14
                node["test_validity"] = 14
                found_positive.add(node.index)
            else:
                node["test_result"] = 0
                node["test_validity"] = incubation_days
    
    # to_quarantine will contain family contacts (quarantine 100%), 
    # possibly_quarantine will contain other contacts, quarantine influenced by contact tracing efficency

    if len(found_positive) > 0:
        to_quarantine = set()
        possibly_quarantine = set()
        # track to max 14 days before
        for i in range(1, min(len(nets) + 1, 15)):
            net = nets[-i]
            for edge in net.es:
                if edge["category"] == "family_contacts" \
                and (edge.source in found_positive or edge.target in found_positive):
                    if edge.source in found_positive:
                        to_quarantine.add(G.vs[edge.target].index)
                    else:
                        to_quarantine.add(G.vs[edge.source].index)
                else:
                    if edge.source in found_positive:
                        possibly_quarantine.add(G.vs[edge.target].index)
                    if edge.target in found_positive:
                        possibly_quarantine.add(G.vs[edge.source].index)
        
        # set diff to remove double contacts
        possibly_quarantine = possibly_quarantine - to_quarantine
        if len(possibly_quarantine) > int(len(possibly_quarantine) * contact_tracking_efficiency):
            possibly_quarantine = random.sample(possibly_quarantine, int(len(possibly_quarantine) * contact_tracking_efficiency))
        else:
            possibly_quarantine = list(possibly_quarantine)
        to_quarantine = list(to_quarantine) + possibly_quarantine

        to_quarantine = [G.vs[i] for i in to_quarantine]

        # put them in quarantine
        # Need test?
        if to_quarantine != list() and to_quarantine != None:
            for node in to_quarantine:
                node["quarantine"] = 14  


'''
def step_vaccine(G, n_vacc, policy_vacc, vacc_pool, agent_status_report):
    """
    Make attributes of nodes consistent
    Make the infection spread across the network
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network
    
    n_vacc: int
        Number of vaccines avaiable that day
    
    policy_vacc: string
        How to distribute the daily avaiable vaccines
        Can be ["Random, Degree Centrality, Betweenness Centrality, Improved Degree Centrality, Improved Betweenness Centrality"]
    
    agent_status_report: list
        Report of nodes status

    Return
    ------
    agent_status_report: list
        Updated report of nodes status

    """

    if len(vacc_pool) > 0:
        
        # random policy
        if policy_vacc == "Random":
            if n_vacc >= len(vacc_pool):
                for index in vacc_pool:
                    agent_status_report[index] = "V"
                    G.nodes[index]["agent_status"] = "V"
            else:
                random_selected_to_vaccinate = random.sample(vacc_pool, n_vacc)
                for index in random_selected_to_vaccinate:
                    agent_status_report[index] = "V"
                    G.nodes[index]["agent_status"] = "V"
        
        # degree policy
        elif policy_vacc == "Degree Centrality":
            degree_results = G.degree(vacc_pool)
            order_by_degree = sorted(degree_results, key=lambda x: x[1], reverse = True)

            for node, value in order_by_degree[:n_vacc]:
                agent_status_report[node] = "V"
                G.nodes[node]["agent_status"] = "V"

        elif policy_vacc == "Improved Degree Centrality":
            Z = G.subgraph(vacc_pool)
            degree_results = Z.degree()
            order_by_degree = sorted(degree_results, key=lambda x: x[1], reverse = True)

            for node, value in order_by_degree[:n_vacc]:
                agent_status_report[node] = "V"
                G.nodes[node]["agent_status"] = "V"

        # Betweenness policy
        elif policy_vacc == "Betweenness Centrality":
            betweenness_results = {node: val for node, val in ig.betweenness_centrality(G).items() if node in vacc_pool}

            order_by_betweenness = sorted(betweenness_results, key=betweenness_results.get, reverse = True)
            
            for node in order_by_betweenness[:n_vacc]:
                agent_status_report[node] = "V"
                G.nodes[node]["agent_status"] = "V"

        elif policy_vacc == "Improved Betweenness Centrality":
            betweenness_results = {node: val for node, val in ig.betweenness_centrality_subset(G, sources = vacc_pool, targets = vacc_pool).items() if node in vacc_pool}

            order_by_betweenness = sorted(betweenness_results, key=betweenness_results.get, reverse = True)
            
            for node in order_by_betweenness[:n_vacc]:
                agent_status_report[node] = "V"
                G.nodes[node]["agent_status"] = "V"
    return agent_status_report  

'''

def step(G, step_index, incubation_days, infection_duration, infection_rate,
         initial_day_restriction, restriction_duration, social_distance_strictness,
         restriction_decreasing, nets, n_test, policy_test, contact_tracking_efficiency):
    """
    Advance the simulation of one step
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network
        
    step_index : int 
        index of the step

    incubation_days: int
        Average number of days where the patient is not infective

    infection_duration: int
        Average total duration of the disease
    
    infection_rate: float
        Value of the infection_rate rate for the disease in the network
    
    initial_day_restriction: int
        Day index from when social distancing measures are applied

    restriction_duration: int
        How many days the social distancing last. Use -1 to make the restriction last till the end of the simulation

    social_distance_strictness: int
        How strict from 0 to 4 the social distancing measures are. 
        Represent the portion of contact that are dropped in the network (0, 25%, 50%, 75%, 100%)
        Note that family contacts are not included in this reduction

    restriction_decreasing: bool
        If the social distancing will decrease the strictness during the restriction_duration period

    nets: list of ig.Graph()
        History of the network

    n_test: int
        Number of avaiable tests

    policy_test: string
        Test strategy
    
    contact_tracking_efficiency: float
        The percentage of contacts successfully traced

    Return
    ------
    report: dict
        Counter of nodes status

    G: ig.Graph()
        The contact network
        
    """
    # generate new edges
    if not restriction_duration:
        if step_index >= initial_day_restriction:
            step_edges(G, 1 - (25 * social_distance_strictness / 100))
        else:
            step_edges(G, 1)
    else:
        if step_index >= initial_day_restriction and step_index < initial_day_restriction + restriction_duration:
            if restriction_decreasing:
                social_distance_strictness = compute_sd_reduction(step_index, initial_day_restriction, restriction_duration, social_distance_strictness)
                step_edges(G, 1 - (25 * social_distance_strictness / 100))
            else:
                step_edges(G, 1 - (25 * social_distance_strictness / 100))
        else:
            step_edges(G, 1)
    
    # spread infection
    step_spread(G, incubation_days, infection_duration, infection_rate)
          
    # make some test on nodes     
    step_test(G, nets, incubation_days, n_test, policy_test, contact_tracking_efficiency)

    '''
    # if avaiable, vaccinate some nodes
    if step_index >= initial_day_vaccination:
        vacc_pool = [i for i in range(len(agent_status_report)) if agent_status_report[i] == "S"]
        agent_status_report = step_vaccine(G, n_vacc, policy_vacc, vacc_pool, agent_status_report)
    '''
    agent_status_report = list()
    for node in G.vs:
        agent_status_report.append(node["agent_status"])

    return Counter(agent_status_report), G  

def compute_sd_reduction(step_index, initial_day_restriction, restriction_duration, social_distance_strictness):
    """
    Calculate the decreased social distance strictness index
    
    Parameters
    ----------
        
    step_index : int 
        index of the step
    
    initial_day_restriction: int
        Day index from when social distancing measures are applied

    restriction_duration: int
        How many days the social distancing last. Use -1 to make the restriction last till the end of the simulation

    social_distance_strictness: int
        How strict from 0 to 4 the social distancing measures are. 
        Represent the portion of contact that are dropped in the network (0, 25%, 50%, 75%, 100%)
        Note that family contacts are not included in this reduction

    Return
    ------
    social_distance_strictness: int
        Updated value for social_distance_strictness
        
    """
    default_days = restriction_duration // social_distance_strictness
    spare_days = restriction_duration - default_days * social_distance_strictness

    strictness_sequence = list()

    for i in range(social_distance_strictness):
        updated_days = default_days
        if spare_days > 0:
            updated_days += 1
            spare_days -= 1
        strictness_sequence.append(updated_days)


    for i in range(1, len(strictness_sequence)):
      strictness_sequence[i] += strictness_sequence[i - 1]


    social_distance_reduction = None

    for i in range(len(strictness_sequence)):
      if step_index - initial_day_restriction < strictness_sequence[i]:
        social_distance_reduction = i
        break

    return social_distance_strictness - social_distance_reduction


import igraph as ig
import numpy as np
from pathlib import Path
import sys, random, time, pickle
try:
    from ctns.generator import generate_network, init_infection
    from ctns.steps import step
    from ctns.utility import dump_network, compute_IR
except ImportError as e:
    from generator import generate_network, init_infection
    from steps import step
    from utility import dump_network, compute_IR

def run_simulation(n_of_families = 500,
    use_steps = True,
    number_of_steps = 150,
    incubation_days = 5,
    infection_duration = 21,
    initial_day_restriction = 50,
    restriction_duration = 21,
    social_distance_strictness = 2,
    restriction_decreasing = True,
    n_initial_infected_nodes = 10,
    R_0 = 2.9,
    n_test = 5,
    policy_test = "Random",
    contact_tracking_efficiency = 0.8,
    use_random_seed = None,
    seed = None,
    dump = False,
    path = None):

    """
    Execute the simulation and dump/return resulting networks

    Parameters
    ----------
    
    n_of_families: int
        Number of families in the network
        
    use_steps : bool 
        Use a fixed number of steps or keep the simulation going untill the spreading is not over

    number_of_steps : int
        Number of simulation step to perform

    incubation_days: int
        Number of days where the patient is not infective

    infection_duration: int
        Total duration of the disease per patient

    initial_day_restriction: int
        Day index from when social distancing measures are applied

    restriction_duration: int
        How many days the social distancing last. Use 0 to make the restriction last till the end of the simulation

    social_distance_strictness: int
        How strict from 0 to 4 the social distancing measures are. 
        Represent the portion of contact that are dropped in the network (0, 25%, 50%, 75%, 100%)
        Note that family contacts are not included in this reduction

    restriction_decreasing: bool
        If the social distancing will decrease the strictness during the restriction_duration period

    n_initial_infected_nodes: int
        Number of nodes that are initially infected

    R_0: float
        The R0 facotr of the disease

    n_test: int
        Number of avaiable tests

    policy_test: string
        Strategy with which test are made. Can be Random, Degree Centrality, Betweenness Centrality

    contact_tracking_efficiency: float
        The percentage of contacts successfully traced back in the past 14 days

    use_random_seed: bool
        Id use or not a fixed random seed

    seed: int
        The random seed value

    dump: bool
        If true dump networks to file, otherwise return the networks list
        NB, if network simulation is too heavy, the dump can crash

    path:string
        The path to the file/folder where the networks will be saved

    Return
    ------

    nets:list
        List of nets of the simulation

    """
    # generate new edges

    if use_random_seed:
        np.random.seed(seed = seed)
        random.seed(seed)
    else:
        np.random.seed(int(time.time()))
        random.seed(time.time())
    
    # check values
    if n_of_families < 10:
        print("Invalid number of families. Use at least 10 families")
        sys.exit()
    if use_steps:
        if number_of_steps < 0:
            print("Invalid number of steps")
            sys.exit()
    if infection_duration < 0:
        print("Invalid infection duration")
        sys.exit()
    if incubation_days < 0 or incubation_days >= infection_duration:
        print("Invalid incubation duration")
        sys.exit()
    if initial_day_restriction < 0 :
        print("Invalid initial day social distancing")
        sys.exit()
    if social_distance_strictness < 0 or social_distance_strictness > 4:
        print("Invalid social distancing value")
        sys.exit()
    if n_initial_infected_nodes < 0 or n_initial_infected_nodes > n_of_families:
        print("Invalid number of initial infected nodes")
        sys.exit()
    if R_0 < 0:
        print("Invalid value of R0")
        sys.exit()
    if n_test < 0:
        print("Invalid number of test per day")
        sys.exit()
    if not (policy_test == "Random" or policy_test == "Degree Centrality" or policy_test == "Betweenness Centrality"):
        print("Invalid test strategy")
        sys.exit()
    if contact_tracking_efficiency < 0 or contact_tracking_efficiency > 1:
        print("Invalid contact tracing efficiency")
        sys.exit()
    if restriction_duration < 0:
        print("Invalid restriction restriction_duration")
        sys.exit()
    if restriction_duration == 0:
        restriction_decreasing = False
        social_distance_strictness = 0
    if social_distance_strictness == 0:
        restriction_decreasing = False
        restriction_duration = 0


    # init network
    G = generate_network(n_of_families)
    infection_rate = compute_IR(G, R_0, infection_duration, incubation_days)
    init_infection(G, n_initial_infected_nodes)
    nets = list()

    if use_steps:
        for sim_index in range (0, number_of_steps):
            report, net= step(G, sim_index, incubation_days, infection_duration, infection_rate,
                             initial_day_restriction, restriction_duration, social_distance_strictness, 
                             restriction_decreasing, nets, n_test, policy_test, contact_tracking_efficiency)
            nets.append(net.copy())
    else:
        exposed = n_initial_infected_nodes
        infected = 0
        sim_index = 0
        while((infected + exposed) != 0):
            report, net= step(G, sim_index, incubation_days, infection_duration, infection_rate,
                             initial_day_restriction, restriction_duration, social_distance_strictness, 
                             restriction_decreasing, nets, n_test, policy_test, contact_tracking_efficiency)
            nets.append(net.copy())
            sim_index += 1
            infected = report["I"]
            exposed = report["E"]

            if infected + exposed == 0:
                break
    if dump:
        try:
            with open(Path(path + ".pickle"), "wb") as f:
                pickle.dump(nets, f)
        except:
            print("Simulation is too big to be dumped, try dumping in separated files. Note that this operation will be slow")
            try:
            	for i in range(len(nets)):
            		dump_network(net, Path(path + "_network{}.pickle".format(i)))
            except:
            	print("Cannot dump network, the single network is too big. Please deactivate the dump option")
            	sys.exit(-1)
    
    print("\n Simulation ended successfully \n You can find the dumped networks in the choosen folder \n")

    return nets

def main():

    n_of_families = None
    use_steps = None
    number_of_steps = None
    incubation_days = None
    infection_duration = None
    initial_day_restriction = None
    restriction_duration = None
    social_distance_strictness = None
    restriction_decreasing = None
    n_initial_infected_nodes = None
    R_0 = None
    n_test = None
    policy_test = None
    contact_tracking_efficiency = None
    use_random_seed = None
    seed = None
    dump = True
    path = None

    user_interaction = int(input("Press 0 to load the default values or 1 to manually input the configuration for the simulation: "))
    # get values from user
    if user_interaction:
        n_of_families = int(input("Please insert the number of families in the simulation: "))
        use_steps = int(input("Press 1 use a fixed number of steps or 0 to run the simulation untill the infection is over: "))
        if use_steps:
            number_of_steps = int(input("Please insert the number of the simulation step: "))
        incubation_days = int(input("Please insert the disease incubation duration: "))
        infection_duration = int(input("Please insert the disease duration: "))
        initial_day_restriction = int(input("Please insert the step index from which the social distance is applied: "))
        restriction_duration = int(input("Please insert the number of days which the social distance last. Insert 0 to make the restriction last for all the simulation: "))
        social_distance_strictness = int(input("Please insert a value between 0 and 4 to set the social distance strictness: "))
        restriction_decreasing = int(input("Press 0 to make the strictness of the social distance decrease during the simulation or 1 to keep it fixed: "))
        n_initial_infected_nodes = int(input("Please insert the number of initial infected individuals: "))
        R_0 = float(input("Please insert the value of R0: "))
        n_test = int(input("Please insert the number of available test per day: "))
        policy_test = input("Please insert strategy with which test are made. Can be Random, Degree Centrality, Betweenness Centrality: ")
        contact_tracking_efficiency = float(input("Please insert a value between 0 and 1 to set the contact tracing efficiency: "))
        use_random_seed = int(input("Press 1 use a fixed a random seed or 0 to pick a random seed: "))
        if use_random_seed:
            seed = int(input("Please insert the random seed: "))
        path = input("Please insert the path with the file to dump. Please omit file type, that will be set automatically: ")

        run_simulation(n_of_families, use_steps, number_of_steps, incubation_days, infection_duration,
            initial_day_restriction, restriction_duration, social_distance_strictness, restriction_decreasing,
            n_initial_infected_nodes, R_0, n_test, policy_test, contact_tracking_efficiency, use_random_seed, seed, dump, path)
    else:
        path = input("Please insert the path with the file to dump. Please omit file type, that will be set automatically: ")
        run_simulation(path = path, dump = True)

if __name__ == "__main__":
    main()


import igraph as ig
import random, math
import numpy as np
try:
    from ctns.utility import fix_distribution_node_number
except ImportError as e:
    from utility import fix_distribution_node_number

def generate_node_list_attribute(G, attribute_name, distribution):
    """
    Add to each node a list named attribute_name containing edge representing the
    possible social contanct to each node.
    This list is used to draw the new edges for the node in the step function
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network

    attribute_name: string
        Name of the node attribute

    distribution: list of int
        Distribution of the "community" dimension
        
    Return
    ------
    None

    """
    # extract node list and randomize
    node_list = list(G.vs)
    random.shuffle(node_list)
    # iterate over community
    for element in distribution:
        community_nodes = list()
        # extract community members
        for i in range(0, element):
            community_nodes.append(node_list.pop())
        # for each node, pick a random number (proportional to sociality) of 
        # community member and create a possible link
        for node in community_nodes:
            tmp = int(math.ceil(len(community_nodes) / 3))
            tmp2 = int(math.ceil(2 * len(community_nodes) / 3))
            if node["sociality"] == "low":
                n_contact = 1 + int(random.random() * (tmp - 1))
            if node["sociality"] == "medium":
                n_contact = tmp + int(random.random() * (tmp2 - tmp))
            if node["sociality"] == "high":
                n_contact = tmp2 + int(random.random() * (len(community_nodes) + 1 - tmp2))

            targets = random.sample(community_nodes, n_contact)
            for target in targets:
                if target != node and not (node.index, target.index) in node[attribute_name]:
                    node[attribute_name].append((node.index, target.index))
                    target[attribute_name].append((target.index, node.index))  

def generate_node_family_attribute(G, distribution):
    """
    Add to each node a list named family_contacts containing edge representing the
    interaction with family and a family ID.
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network

    distribution: list of int
        Distribution of the "community" dimension
        
    Return
    ------
    None

    """
    node_counter = 0;
    for i in range(len(distribution)):
        family_nodes = list(range(node_counter, node_counter + distribution[i]))
        node_counter = node_counter + distribution[i]
        # for ech node in the family, generate a complete graph with other family members
        for source in family_nodes:
            G.vs[source]["family_id"] = i
            for target in family_nodes:
                if source != target:
                    G.vs[source]["family_contacts"].append((source, target))  

def generate_network(n_of_families):
    """
    Generte contact network nodes
    
    Parameters
    ----------
    n_of_families: int
        Number of families in the network
        
    Return
    ------
    G: ig.Graph()
        The contact network

    """

    #DISTRIBUTIONS
    # family distribution
    family_distribution = np.random.normal(2, 2, n_of_families).round().astype(int)
    family_distribution = [x if x > 1 else 1 for x in family_distribution]

    number_of_nodes = np.sum(family_distribution)


    # frequent contact distribution
    frequent_contact_distribution = np.random.normal(12, 3, n_of_families).round().astype(int)
    frequent_contact_distribution = list(frequent_contact_distribution[frequent_contact_distribution >= 0])

    # refine to have sum = node number
    frequent_contact_distribution = fix_distribution_node_number(frequent_contact_distribution, number_of_nodes)


    # occasional contact distribution
    occasional_contact_distribution = np.random.normal(24, 6, n_of_families).round().astype(int)
    occasional_contact_distribution = list(occasional_contact_distribution[occasional_contact_distribution >= 0])

    # refine to have sum = node number
    occasional_contact_distribution = fix_distribution_node_number(occasional_contact_distribution, number_of_nodes)

    # random edge distribution -> see step function

    # GENERATE NODES
    G = ig.Graph()
    G.add_vertices(number_of_nodes)
    for node in G.vs:
        node["sex"] = np.random.choice(["man", "woman"], p = [0.487, 0.513])
        # age is the represented by the lower bound, so if age is 20, the person has age [20-29]
        node["age"] = np.random.choice([0, 10, 20, 30, 40, 50, 60, 70, 80, 90], 
            p = [0.084, 0.096, 0.102, 0.117, 0.153, 0.155, 0.122, 0.099, 0.059, 0.013]) 

        node["family_id"] = -1
        node["family_contacts"] = list()
        node["frequent_contacts"] = list()
        node["occasional_contacts"] = list()

        # setting sociality
        if node["age"] < 6:
            node["sociality"] = "low"
            node["pre_existing_conditions"] = 0
        elif node["age"] > 70:
            node["sociality"] = np.random.choice(
                ['low', 'medium', 'high'], p=[0.75, 0.23, 0.02])
            node["pre_existing_conditions"] = np.random.choice(
                [0, 1, 2, 3], p=[0.1, 0.4, 0.3, 0.2])
        else:
            node["sociality"] = np.random.choice(
                ['low', 'medium', 'high'], p=[0.6, 0.3, 0.1])
            node["pre_existing_conditions"] = np.random.choice(
                [0, 1, 2, 3], p=[0.6, 0.2, 0.1, 0.1])
        
        if node["age"] < 10:
            node["death_rate"] = 0.002
        elif node["age"] < 20:
            node["death_rate"] = 0.001
        elif node["age"] < 30:
            node["death_rate"] = 0.001
        elif node["age"] < 40:
            node["death_rate"] = 0.004
        elif node["age"] < 50:
            node["death_rate"] = 0.009
        elif node["age"] < 60:
            node["death_rate"] = 0.026
        elif node["age"] < 70:
            node["death_rate"] = 0.10
        elif node["age"] < 80:
            node["death_rate"] = 0.249
        elif node["age"] < 90:
            node["death_rate"] = 0.308
        elif node["age"] < 100:
            node["death_rate"] = 0.261

        node["agent_status"] = 'S'
        node["infected"] = False
        node["days_from_infection"] = 0
        node["probability_of_being_infected"] = 0.0
        node["quarantine"] = 0
        node["test_validity"] = 0
        node["test_result"] = -1
        node["symptoms"] = list()
        node["needs_IC"] = False

    # GENERATE NODE LIST ATTRIBUTES
    # generate families contact attribute in nodes
    generate_node_family_attribute(G, family_distribution)
    # generate frequent contact attribute in nodes
    generate_node_list_attribute(G, "frequent_contacts", frequent_contact_distribution)
    # generate occasional contact attribute in nodes
    generate_node_list_attribute(G, "occasional_contacts", occasional_contact_distribution)

    return G  

def init_infection(G, n_initial_infected_nodes):
    """
    Make random nodes infected in the network
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network

    n_initial_infected_nodes: int
        Number of infected nodes to put on the network

    Return
    ------
    None

    """
    for node in random.sample(list(G.vs), n_initial_infected_nodes):
        node["agent_status"] = "E"
        node["infected"] = True
        node["days_from_infection"] = 1


import igraph as ig
import random
import numpy as np
#import matplotlib.pyplot as plt
from collections import Counter
try:
    from ctns.steps import step
except ImportError as e:
    from steps import step

def fix_distribution_node_number(distribution, n_nodes):
    """
    Make the sum of the elements in the distribution be equal to the number of nodes
    
    Parameters
    ----------
    distribution: list of int
        Distribution of int

    n_nodes: int
        Number of nodes in The contact network
        
    Return
    ------
    distribution: list of int
        Distribution where the sum of elements is equal to n_nodes

    """
    refined_distribution = list()
    while True:
        refined_distribution.append(distribution.pop())
        if (np.sum(refined_distribution) > n_nodes):
            refined_distribution.pop()
            break
    refined_distribution.append(int(n_nodes - np.sum(refined_distribution)))
    return refined_distribution  

def reset_network(G):
    """
    Reset network status, removing all edges and setting all the infection 
    related attributes of each node to the default value
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network
        
    Return
    ------
    None

    """
    G.delete_edges(list(G.es))
    for node in G.vs:
        node["agent_status"] = 'S'
        node["infected"] = False
        node["days_from_infection"] = 0
        node["probability_of_being_infected"] = 0.0
        node["quarantine"] = 0
        node["test_validity"] = 0
        node["test_result"] = -1
        node["symptoms"] = list()  

def dump_network(G, path):
    """
    Dump the contact network to a file
    N.B.: if you need to reload the network after dump, use only pickle, not json 
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network

    path: string
        The file path
    Return
    ------
    None

    """
    G.write_picklez(path)

def compute_IR(G, R_0, infection_duration, incubation_days):
    """
    Compute the infection rate of the disease in the network.
    The factor is computed as R_0 / (average_weighted_degree * (infection_duration - incubation_days))
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network

    R_0: float
        R_0 of the disease

    infection_duration: int
        Average total duration of the disease

    incubation_days: int
        Average number of days where the patient is not infective

    Return
    ------
    infection_rate: float
        The infection rate for the network

    """
    avr_deg = list()
    # compute average weighted degree on 20 steps
    for i in range (20):
        step(G, i, 0, 0, 0, 0, 0, 0, False, list(), 0, "Random", 0)

        degrees = G.strength(list(range(len(G.vs))), weights = "weight")
        avr_deg.append(sum(degrees) / len(degrees))
    #reset network status
    reset_network(G)
    return R_0 /((infection_duration - incubation_days) * (sum(avr_deg) / len(avr_deg)))

'''

def plot_degree_dist(G, node_sociality = None, edge_category = None, title = None):
    """
    Print degree probability distribution of the network.
    node_sociality and edge_category can be used to filter only certain types of
    nodes and edges
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network

    node_sociality: string
        Consider only nodes that have sociality == node_sociality

    edge_category: string
        Consider only edges that have category == edge_category

    title: string
        The title of the plot

    Return
    ------
    None

    """
    #plt.figure(figsize = (8, 6), dpi = 300)
    if edge_category != None:
        G = G.copy()
        toRemove = []
        for edge in G.es:
            if edge["category"] != edge_category:
                toRemove.append(edge.index)
        G.delete_edges(toRemove)


    if node_sociality == None:
        degs = G.degree()
    else:
        selected_nodes = list()
        for node in G.vs:
            if node['sociality'] == node_sociality:
                selected_nodes.append(node)
        degs = G.degree(selected_nodes) 

    conuted_degs = Counter(degs)
    for key in conuted_degs.keys():
        conuted_degs[key] = conuted_degs[key] / len(G.vs())
        plt.scatter(key, conuted_degs[key], marker = '.', color = "red", s = 10)
    plt.xticks(fontsize = 12)
    plt.yticks(fontsize = 12)
    plt.xlabel("Degree", fontsize = 15)
    plt.ylabel("Probability", fontsize = 15)
    if title == None:
        plt.title("Degree distribution", fontsize = 20)
    else:
        plt.title(title, fontsize = 20)
    plt.tight_layout()
    plt.show()  

# Print degree distribution
def print_degree_summary(G):
    """
    Print degree probability distribution dummary of the network.
    
    Parameters
    ----------
    G: ig.Graph()
        The contact network

    Return
    ------
    None

    """
    plot_degree_dist(G, title = "Degree distribution")
    plot_degree_dist(G, node_sociality = "low", title = "Degree distribution low sociality")
    plot_degree_dist(G, node_sociality = "medium", title = "Degree distribution medium sociality")
    plot_degree_dist(G, node_sociality = "high", title = "Degree distribution high sociality")

    plot_degree_dist(G, edge_category = "family_contacts", title = "Degree distribution family contacts")
    plot_degree_dist(G, edge_category = "frequent_contacts", title = "Degree distribution frequent contacts")
    plot_degree_dist(G, edge_category = "occasional_contacts", title = "Degree distribution occasional contacts")
    plot_degree_dist(G, edge_category = "random_contacts", title = "Degree distribution random contacts")

'''